// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests sync.Once fast-path scalability under contention.
// A set of goroutines repeatedly execute sync.Once.Do() on a single object.
// Expected to scale almost linearly.

package main

import (
	"container/vector"
	"runtime"
	"sync"
)

func benchmarkOnce() {
	procs := runtime.GOMAXPROCS(-1)
	compl := make(chan uint64, procs)
	once := &sync.Once{}
	for p := 0; p != procs; p += 1 {
		go func() {
			f := func() {}
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				once.Do(f)
			}
			compl <- work
		}()
	}
	for p := 0; p != procs; p += 1 {
		totalWork += <-compl
	}
}

func SuiteOnce(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"once", benchmarkOnce})
}
