// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests channel scalability in single-producer/single-consumer mode.
// slack variations create more goroutines than there are worker threads.
// Is not expected to scale.

package main

import (
	"container/vector"
	"runtime"
)

func benchmarkPingPong(slack, chanSize int) {
	procs := runtime.GOMAXPROCS(-1) * slack / 2
	if procs == 0 {
		procs = 1
	}
	compl := make(chan uint64, 2*procs)
	for p := 0; p != procs; p += 1 {
		ping := make(chan int, chanSize)
		pong := make(chan int, chanSize)
		go func() {
			pong <- 0
			for true {
				v := <-ping
				if v == 1 {
					break
				}
				pong <- 0
			}
			compl <- 0
		}()
		go func() {
			work := uint64(0)
			for benchmarking {
				work += 1
				<-pong
				ping <- 0
			}
			<-pong
			ping <- 1
			compl <- work
		}()
	}
	for p := 0; p != 2*procs; p += 1 {
		totalWork += <-compl
	}
}

func SuitePingPong(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"pingpong-unbuf", func() {
		benchmarkPingPong(1, 0)
	}})
	benchmarks.Push(&Benchmark{"pingpong-unbuf-slack", func() {
		benchmarkPingPong(4, 0)
	}})
	benchmarks.Push(&Benchmark{"pingpong-buf", func() {
		benchmarkPingPong(1, 4)
	}})
	benchmarks.Push(&Benchmark{"pingpong-buf-slack", func() {
		benchmarkPingPong(4, 4)
	}})
}
