// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Tests performance/scalability of sync.WaitGroup.
// waitgroup-distr tests scalability of the object without contention.
//	Expected to scale almost linearly.
// waitgroup-centr tests scalability of the object under contention.
//	work variation executes some amount of local work between
//	WaitGroup.Add() and WaitGroup.Done().
//	Is not expected to scale.

package main

import (
	"container/vector"
	"runtime"
	"sync"
	"sync/atomic"
	"math"
)

func benchmarkWaitGroupDistrImpl(wg *sync.WaitGroup, level, spread int) {
	if level > 0 {
		wg0 := &sync.WaitGroup{}
		wg0.Add(spread)
		for i := 0; i != spread; i += 1 {
			go benchmarkWaitGroupDistrImpl(wg0, level-1, spread)
		}
		wg0.Wait()
	}
	wg.Done()
}

func benchmarkWaitGroupDistr() {
	spread := 4
	level := 7
	count := uint64(0)
	for i := 0; i <= level; i += 1 {
		count += uint64(math.Pow(float64(spread), float64(i)))
	}
	for benchmarking {
		wg := &sync.WaitGroup{}
		wg.Add(1)
		benchmarkWaitGroupDistrImpl(wg, level, spread)
		totalWork += count
		runtime.Gosched()
	}
}

func benchmarkWaitGroupCentr(localWork int) {
	procs := runtime.GOMAXPROCS(-1)
	wg := &sync.WaitGroup{}
	wg.Add(procs)
	for p := 0; p != procs; p += 1 {
		go func() {
			work := uint64(0)
			for benchmarking {
				work += 1
				if (work % 1000) == 0 {
					runtime.Gosched()
				}
				wg.Add(1)
				for i := 0; i != localWork; i += 1 {
					work *= 2
					work /= 2
				}
				wg.Done()
			}
			atomic.AddUint64(&totalWork, work)
			wg.Done()
		}()
	}
	wg.Wait()
}

func SuiteWaitGroup(benchmarks *vector.Vector) {
	benchmarks.Push(&Benchmark{"waitgroup-distr", func() {
		benchmarkWaitGroupDistr()
	}})
	benchmarks.Push(&Benchmark{"waitgroup-centr", func() {
		benchmarkWaitGroupCentr(0)
	}})
	benchmarks.Push(&Benchmark{"waitgroup-centr-work", func() {
		benchmarkWaitGroupCentr(100)
	}})
}
